import json

import click
import curies
import rdflib
from rdflib import Graph, term, XSD, URIRef, Namespace


@click.command()
@click.option(
    "-i",
    "--input-ttl",
    required=True,
    type=click.Path(exists=True, dir_okay=False),
)
@click.option(
    "-o",
    "--output-ttl",
    required=True,
    type=click.Path(dir_okay=False),
)
@click.option(
    "-p",
    "--jsonld-context-jsons",
    multiple=True,
    required=True,
    type=click.Path(dir_okay=False),
)
@click.option(
    "-p",
    "--emsl-uuid-replacement",
    required=True,
    help="A prefix, from one of the jsonld-context-jsons, that should replace the UUID prefix used in some emsl_biosample_identifiers values",
)
def expand_curies(input_ttl, output_ttl, jsonld_context_jsons, emsl_uuid_replacement):
    """Expand CURIE literals in an RDF graph to their full URI references."""

    graph = Graph()

    for_chaining = []
    for jsonld_context_file in jsonld_context_jsons:
        print(f"Loading prefixes from {jsonld_context_file}")
        with open(jsonld_context_file) as json_file:
            temp_dict = json.load(json_file)
        if '@context' in temp_dict and '@vocab' in temp_dict['@context']:  # todo make a list of excluded keys
            del temp_dict['@context']['@vocab']
        temp_context = curies.load_jsonld_context(temp_dict)
        for_chaining.append(temp_context)

    converter = curies.chain(for_chaining)

    for prefix, uri_prefix in converter.prefix_map.items():
        graph.bind(prefix, rdflib.Namespace(uri_prefix), override=False, replace=False)

    print(f"Loading {input_ttl}")
    graph.parse(input_ttl, format="ttl")
    print(f"Loaded {input_ttl}")

    print("Iterating over triples")
    for s, p, o in graph:
        if o.__class__ == term.Literal and o.datatype == XSD.anyURI:
            o_str = str(o)
            curie_parts = o_str.split(":")
            current_prefix = curie_parts[0]
            current_suffix = curie_parts[1]
            as_true_curie = None
            if current_prefix == "UUID" and current_suffix:
                if str(p) == "https://w3id.org/nmdc/emsl_biosample_identifiers":
                    as_true_curie = (
                        URIRef(graph.namespace_manager.expand_curie(
                            f"{emsl_uuid_replacement}:{current_suffix}")))
            else:
                try:
                    as_true_curie = URIRef(graph.namespace_manager.expand_curie(o_str))
                except Exception as e:
                    if current_prefix.isupper():
                        as_true_curie = URIRef(
                            graph.namespace_manager.expand_curie(f"{current_prefix.lower()}:{current_suffix}"))
                    elif current_prefix.islower():
                        as_true_curie = URIRef(
                            graph.namespace_manager.expand_curie(f"{current_prefix.upper()}:{current_suffix}"))
                    else:
                        print(e)
            graph.add((s, p, as_true_curie))
            graph.remove((s, p, o))

    print(f"Serializing to {output_ttl}")
    graph.serialize(
        destination=output_ttl
    )

    click.echo("Expanded CURIE literals in RDF graph.")


if __name__ == "__main__":
    expand_curies()



import logging
import os
import pprint

import click
import click_log
import requests
from dotenv import load_dotenv
from linkml_runtime import SchemaView
from linkml_runtime.dumpers import yaml_dumper
from pymongo import MongoClient
from pymongo.errors import OperationFailure

logger = logging.getLogger(__name__)
click_log.basic_config(logger)


# todo: add a collection report only mode
# todo balance extensiveness of reporting with speed of execution, even in a degraded state?


# Get NERSC sshproxy app. See https://docs.nersc.gov/connect/mfa/#sshproxy.
#   Documentation may not be completely up-to-date.
# scp <NERSC_USER_NAME>@dtn01.nersc.gov:/global/cfs/cdirs/mfa/NERSC-MFA/sshproxy.sh .
#   Requires that you have a NERSC account. See ???
#   May require entering your NERSC password and MFA token
# <PATH_TO>/sshproxy.sh -u <NERSC_USER_NAME>
#   Make a note of the location where your key has been saved. Default = ~/.ssh/nersc
# ssh -i ~/.ssh/nersc -L27777:mongo-loadbalancer.nmdc.production.svc.spin.nersc.org:27017 -o ServerAliveInterval=60 <NERSC_USER_NAME>@dtn01.nersc.gov
#  You can use a local port other than 27777 if you want. Just be consistent in the rest of this setup.


def set_arithmetic(set1, set2, set1_name='set 1 only', set2_name='set 2 only'):
    set1_only = set1 - set2
    set2_only = set2 - set1
    intersection = set1.intersection(set2)
    temp = {
        f"{set1_name} only": list(set1_only),
        f"{set2_name} only": list(set2_only),
        'intersection': list(intersection),  # todo should the name of the intersection be a parameter?
    }
    return temp


class FastAPIClient:
    def __init__(self, base_url):
        self.base_url = base_url

    def get_collection_stats_from_fastapi(self):
        pass

    def get_docs_from_fastapi(self, method, endpoint, params=None, data=None):
        url = f"{self.base_url}/{endpoint}"
        try:
            response = requests.request(method, url, params=params, json=data)
            response.raise_for_status()
            return response.json()
        except Exception as e:
            logger.warning(f"{e}")

    def get_paginated_data(self, endpoint, params, max_docs=None, results_key='resources',
                           continuation_key='next_page_token', continuation_parameter='page_token'):
        params = params or {}
        data = []

        while True:
            response = self.get_docs_from_fastapi('GET', endpoint, params=params)
            if response and results_key in response:
                temp = response[results_key]
                data.extend(temp)
                data_len = len(data)
                logger.info(f"Retrieved {data_len} entries out of {max_docs} from {endpoint}")
                if data_len >= max_docs:
                    break

                if continuation_key in response:
                    params[continuation_parameter] = response[continuation_key]
                else:
                    break
            else:
                # logger.warning(f"FastAPI request to {endpoint} failed. Might work as a pymongo query.")
                break

        return data


class PyMongoClient:

    def __init__(self, env_file, mongo_db_name, mongo_host, mongo_port, admin_db, auth_mechanism='SCRAM-SHA-256',
                 direct_connection=True):
        self.admin_db = admin_db
        self.auth_mechanism = auth_mechanism
        self.direct_connection = direct_connection
        self.env_file = env_file
        self.mongo_db_name = mongo_db_name
        self.mongo_host = mongo_host
        self.mongo_port = mongo_port
        self.selected_collections = []

        logger.debug("Attempting to create a MongoDB database object")
        self.db = self.create_database_obj()
        logger.debug(pprint.pformat(self.__dict__))
        logger.info("MongoDB database object created. Now attempting to get collection names.")
        self.collections = self.get_collection_names_from_pymongo()
        logger.info(f"Found {len(self.collections)} collections in MongoDB database {self.mongo_db_name}")

    def create_database_obj(self):
        # Load MongoDB credentials from .env file
        load_dotenv(self.env_file)
        logger.info(f"loaded {self.env_file}")
        mongo_pw = os.getenv('SOURCE_MONGO_PASS')
        mongo_user = os.getenv('SOURCE_MONGO_USER')
        logger.info(f"{mongo_user = }")

        client = MongoClient(host=self.mongo_host,
                             port=self.mongo_port,
                             username=mongo_user,
                             password=mongo_pw,
                             authSource=self.admin_db,
                             authMechanism=self.auth_mechanism,
                             directConnection=self.direct_connection
                             )

        db = client[self.mongo_db_name]

        return db

    def get_collection_names_from_pymongo(self):

        collections = self.db.list_collections()

        filtered_collections = []

        for collection_info in collections:
            if collection_info["type"] == "view":
                logger.warning(f"Collection '{collection_info['name']}' is a view.")
            else:
                logger.debug(f"Collection '{collection_info['name']}' is a collection.")
                collection_name = collection_info['name']
                logger.debug(f"Checking permissions for collection '{collection_name}'...")
                collection = self.db[collection_name]
                try:
                    document = collection.find_one()  # max_time_ms=1000 ?
                    if document is not None:
                        logger.debug(f"You have read permission for collection '{collection_name}'.")
                        filtered_collections.append(collection_name)
                except OperationFailure as e:
                    if "not authorized" in str(e):
                        logger.warning(f"You do not have read permission for collection '{collection_name}'.")

        return filtered_collections

    def get_collection_stats(self):
        selected_stats = {}
        for collection_name in self.selected_collections:
            for_selected_stats = {}

            collection_stats = self.db.command("collstats", collection_name)

            for_selected_stats["document_count"] = collection_stats["count"]
            for_selected_stats["size_in_bytes"] = collection_stats["size"]
            for_selected_stats["storageSize"] = collection_stats["storageSize"]
            for_selected_stats["avg_bytes_per_doc"] = collection_stats["size"] / collection_stats["count"]

            selected_stats[collection_name] = for_selected_stats

        return selected_stats

    def get_docs_from_pymongo(self, collection_name, max_docs_per_coll):
        collection = self.db[collection_name]

        if collection is None:
            logger.info(f"Could not find collection {collection_name}")

        documents = collection.find().limit(max_docs_per_coll)
        doc_list = list(documents)

        prob_key = "_id"
        for doc in doc_list:
            if prob_key in doc:
                del doc[prob_key]

        return doc_list


class ViewHelper:
    def __init__(self, schema_path):
        self.view = SchemaView(schema_path)

    def get_class_slots(self, class_name, include_scalars=False):
        class_slot_objs = self.view.class_induced_slots(class_name)
        class_slot_obj_dict = {s.name: s for s in class_slot_objs}
        class_slot_names = [s.name for s in class_slot_objs]
        acceptable_slots = []
        for i in class_slot_names:
            slot_obj = class_slot_obj_dict[i]
            # logger.info(f"{i = } {slot_obj.multivalued = }")
            if slot_obj.multivalued or include_scalars:
                # logger.info(f"Adding {i} to ")
                acceptable_slots.append(i)
            else:
                logger.warning(
                    f"Skipping {class_name} slot {i} because it is not multivalued and include_scalars is False.")
        class_slot_names.sort()
        return acceptable_slots


@click.command()
@click_log.simple_verbosity_option(logger)
@click.option('--admin-db', default="admin", show_default=True, help='MongoDB authentication source')
@click.option('--env-file', type=click.Path(exists=True), default='local/.env',
              show_default=True, help='Path to .env file')
@click.option('--max-docs-per-coll', default=100,
              show_default=True, help='Maximum number of documents to retrieve per collection')  # was 100_000
@click.option('--mongo-db-name', default="nmdc", show_default=True, help='MongoDB database name')
@click.option('--mongo-host', default="localhost", show_default=True,
              help='MongoDB host name/address. Use "localhost" if using an ssh tunnel to ' +
                   'mongo-loadbalancer.nmdc.production.svc.spin.nersc.org ' +
                   'or mongo-loadbalancer.nmdc-dev.production.svc.spin.nersc.org')
@click.option('--mongo-port', default=27777, show_default=True, help='MongoDB port')
@click.option('--output-yaml', required=True, type=click.Path(),
              show_default=True, help="Output file.")
@click.option('--root-class', default="Database",
              show_default=True, help='Schema class that corresponds to a Mongo Database')
@click.option('--schema-file', type=click.Path(exists=True), default='src/schema/nmdc.yaml',
              show_default=True,
              help='Path to root YAML file in the nmdc-schema. ' +
                   'For determining structure only. No validation in this step.')
@click.option('--selected-collections', multiple=True, show_default=True, help='MongoDB collection name')
@click.option('--page-size', default=10,
              show_default=True,
              help='Number of documents retrieved per FastAPI GET request')  # how many is too many? for what reason? seems like there's some overhead
@click.option('--client-base-url', default="https://api.microbiomedata.org",
              show_default=True,
              help='HTTP(S) path to the FastAPI server. ' +
                   'https://api-dev.microbiomedata.org/docs or https://api.microbiomedata.org/docs')
@click.option('--endpoint-prefix', default="nmdcschema",
              show_default=True, help='FastAPI path component between the URL and the endpoint name')
@click.option('--collection-check/--skip-collection-check', default=True)
def cli(
        admin_db,
        env_file,
        mongo_db_name,
        mongo_host,
        mongo_port,
        root_class,
        schema_file,
        selected_collections,
        client_base_url,
        endpoint_prefix,
        max_docs_per_coll,
        page_size,
        output_yaml,
        collection_check,

):
    # selected_collections = []
    est_doc_count = 0
    if collection_check:
        nmdc_pymongo_client = PyMongoClient(
            admin_db=admin_db,
            auth_mechanism='SCRAM-SHA-256',
            direct_connection=True,
            env_file=env_file,
            mongo_db_name=mongo_db_name,
            mongo_host=mongo_host,
            mongo_port=mongo_port,
        )
        # logger.info(f"{nmdc_pymongo_client.collections = }")

        nmdc_helper = ViewHelper(schema_file)

        # logger.info(f"{nmdc_helper.view.schema.name = }")

        root_class_slots = nmdc_helper.get_class_slots(root_class)

        # logger.info(f"{root_class_slots = }")

        schema_vs_mongo_collections = set_arithmetic(set(root_class_slots), set(nmdc_pymongo_client.collections),
                                                     set1_name='schema',
                                                     set2_name='mongo')

        logger.info(f"schema_vs_mongo_collections = ")
        logger.info(pprint.pformat(schema_vs_mongo_collections))

        available_selected_collections = []
        if len(selected_collections) > 0:
            available_vs_selected_collections = set_arithmetic(set(schema_vs_mongo_collections['intersection']),
                                                               set(selected_collections), set1_name='available',
                                                               set2_name='selected')
            logger.debug(f"available_vs_selected_collections = ")
            logger.debug(pprint.pformat(available_vs_selected_collections))
            available_selected_collections = available_vs_selected_collections['intersection']
            if available_vs_selected_collections['selected only']:
                logger.warning(
                    f"Some requested collections are not available: {available_vs_selected_collections['selected only']}")
        else:
            available_selected_collections = schema_vs_mongo_collections['intersection']
        available_selected_collections.sort()
        logger.info(f"available_selected_collections = ")
        logger.info(pprint.pformat(available_selected_collections))

        selected_collections = available_selected_collections
    else:
        selected_collections = selected_collections
        logger.info(f"{selected_collections = }")

    # collection_stats = nmdc_pymongo_client.get_collection_stats()

    nmdc_fastapi_client = FastAPIClient(client_base_url)

    if max_docs_per_coll < page_size:
        page_size = max_docs_per_coll

    nmdc_database_object = {}
    for current_collection in selected_collections:

        if collection_check:
            logger.info(f"Attempting to get collection stats from {current_collection}")

            current_coll_obj = nmdc_pymongo_client.db[current_collection]
            est_doc_count = current_coll_obj.estimated_document_count()

            logger.info(
                f"estimated_document_count = {est_doc_count}")  # it would also be nice to report collection size or avg size/doc but I haven't figured how to do that quickly yet

        #     # collection_stats = current_coll_obj.estimated_document_count()
        #     collection_stats = current_coll_obj.stats()
        #
        #     # collection = db[collection_name]
        #     #
        #     # # Get statistics for the collection
        #     # collection_stats = collection.stats()

        endpoint_name = f"{endpoint_prefix}/{current_collection}"
        params_string = {
            "max_page_size": page_size
        }
        max_docs = max_docs_per_coll

        if max_docs_per_coll > est_doc_count:
            max_docs = est_doc_count

        logger.info(
            f"Attempting to get {max_docs} documents from {endpoint_name} in pages of {page_size}.")

        paginated_data = nmdc_fastapi_client.get_paginated_data(endpoint=endpoint_name, params=params_string,
                                                                max_docs=max_docs)

        if paginated_data:
            nmdc_database_object[current_collection] = paginated_data
        else:  # todo needs safer programming like try/except
            logger.warning(f"FastAPI request to {endpoint_name} appears to have failed. Trying as a PyMongo query.")
            direct_data_all = nmdc_pymongo_client.get_docs_from_pymongo(current_collection, max_docs)
            if direct_data_all:
                logger.info(
                    f"Successfully retrieved {len(direct_data_all)} documents from {current_collection} via PyMongo")
                nmdc_database_object[current_collection] = direct_data_all

    logger.info(f"Writing {output_yaml}")
    yaml_dumper.dump(nmdc_database_object, output_yaml)


if __name__ == '__main__':
    cli()



from linkml_runtime.dumpers import yaml_dumper

from nmdc_schema.nmdc_schema_accepting_legacy_ids import Biosample, ControlledIdentifiedTermValue, OntologyClass


def do_test():
    bs = Biosample(id='xxx',
                   part_of='gold:Gs0114663',
                   env_medium=ControlledIdentifiedTermValue(
                       term=OntologyClass(id='ENVO:00010483', name='freshwater environment')),
                   env_broad_scale=ControlledIdentifiedTermValue(
                       term=OntologyClass(id='ENVO:00010483', name='freshwater environment')),
                   env_local_scale=ControlledIdentifiedTermValue(
                       term=OntologyClass(id='ENVO:00010483', name='freshwater environment')),
                   )

    print(yaml_dumper.dumps(bs))



import logging
import re
from typing import Dict, List

import click
import click_log
import yaml
from linkml_runtime import SchemaView

logger = logging.getLogger(__name__)
click_log.basic_config(logger)

# todo: log before and after states of migration

doi_url_pattern = r'^https?:\/\/[a-zA-Z\.]+\/10\.'
curie_pattern = r'^[a-zA-Z_][a-zA-Z0-9_-]*:[a-zA-Z0-9_][a-zA-Z0-9_.-]*$'


class MigratorBase:
    """Base class containing properties and methods related to migrating data between schema versions."""

    def __init__(self):
        self.forced_prefix = None

        # Define the "agenda" of transformations that constitute this migration.
        #
        # Note: This is a dictionary that maps a given collection to a list of "transformation" functions.
        #       Each key is a collection name, and each value is a list. Each element of the list is a
        #       so-called "transformation" function. A "transformation" function is a function that
        #       transforms something from one schema version to another. In this case, the "something"
        #       is a dictionary representing a single document from the specified collection.
        #
        # Note: This dictionary is empty here. It will be populated within the "constructor" functions
        #       of the migration-specific classes (i.e. the classes that inherit from this base class).
        #
        self.agenda: Dict[str, List[callable]] = dict()

    def get_transformers_for(self, collection_name: str) -> List[callable]:
        """Returns the list of transformers defined for the specified collection."""
        return self.agenda.get(collection_name, [])

    def check_and_normalize_one_curie(self, curie_string):
        if not self.is_valid_curie(curie_string):
            curie_string = self.normalize_curie(curie_string)

        return curie_string

    def is_valid_curie(self, curie_string):
        if curie_string == "None":
            pass
            # logger.info("curie_string = 'None'")  # at what point do these get converted from None values to 'None' strings?
        elif '\n' in curie_string or '\r' in curie_string:
            pass
            # logger.info("curie_string contains newline characters")
        else:
            match = re.match(curie_pattern, curie_string)
            return match is not None

    def normalize_curie(self, curie_string):
        # Remove any characters that are not allowed in a CURIE
        # todo add logging!
        curie_cleaned = re.sub(r'[^a-zA-Z0-9:_./-]', '', curie_string)

        # Add a user-specified prefix if no prefix is present
        if ':' not in curie_cleaned:
            curie_normalized: str = f'{self.forced_prefix}:{curie_cleaned}'
        else:
            curie_normalized = curie_cleaned

        return curie_normalized

    def fix_curies_recursively_by_key(self, o, keys_to_migrate, migration_eligible=False):
        if isinstance(o, dict):
            migration_eligible = False
            for k, v in o.items():
                o[k] = self.fix_curies_recursively_by_key(
                    v, keys_to_migrate, migration_eligible or k in keys_to_migrate
                )
            return o
        elif isinstance(o, list):
            return [
                self.fix_curies_recursively_by_key(
                    v, keys_to_migrate, migration_eligible)
                for v in o
            ]
        else:
            return self.check_and_normalize_one_curie(o) if migration_eligible else o


def load_yaml_file(filename):
    """Loads a YAML file into a Python dict."""
    with open(filename, "r") as f:
        data = yaml.safe_load(f)
    return data


class Migrator_from_A_B_C_to_X_Y_Z(MigratorBase):
    """
    Migrates data from schema A.B.C to X.Y.Z

    TUTORIAL: This is an example of a "migrator" class.
              It was designed for use during developer training and
              to serve as a template for production "migrator" classes.
    """

    def __init__(self) -> None:
        """
        TUTORIAL: This is the "constructor" function of the class.
                  As is true about the "constructor" function of any class,
                  it runs whenever that class is instantiated, and its job
                  is to initialize the newly-created instance of that class.

                  When this "constructor" function runs, it does two things:
                  1. It invokes the base class's "constructor" function; and
                  2. It populates a dictionary that keeps track of which
                     transformation functions will be applied to objects
                     from which collections. You can think of this as the
                     "agenda", "itinerary", "plan", or "registry" of
                     transformations that make up this migration. As
                     security guards at marathons say, "If it ain't
                     part of the plan, it ain't gonna be ran."

             -->  As part of creating a new "migration" class, you will
                  populate its "agenda."
        """

        # Invoke the base class's "constructor" function.
        super().__init__()

        # Populate the "collection-to-transformers" map for this specific migration.
        self.agenda = dict(
            study_set=[self.allow_multiple_names],
        )

    def allow_multiple_names(self, study: dict) -> dict:
        """
        TUTORIAL: This is an example "transformation" function within the class.
                  Its jobs is to transform a dictionary that conforms to the
                  initial schema version (in this case, version "A.B.C"), into one
                  that conforms to the target schema version (in this case,
                  version "X.Y.Z"). That might involve adding a field,
                  converting a string into a list of strings, etc.

                  When this "transformation" function runs, it does three things:
                  1. It accepts a single dictionary as a parameter.
                  2. It transforms that dictionary.
                  3. It returns the transformed dictionary.

              --> As part of creating a new "migration" class, you will
                  typically implement one or more "transformation" functions.
                  You will also add them to the "agenda" of the class.
        """

        # optional log message
        logger.info(f"Transforming study having id: {study['id']}")

        # Transform the dictionary.
        #
        # TUTORIAL: In this example scenario, I am pretending that:
        #           1. In schema version "A.B.C", the `Study` class has a `name` slot,
        #              which contain a single string.
        #           2. In schema version "X.Y.Z", the `Study` class no longer has that
        #              `name` slot. Instead, it has a `names` slot, which contains
        #              a list of strings.
        #
        original_name = study["name"]  # preserve the original value
        # create a new key, whose value is an empty list of names
        study["names"] = []
        # add the original value to that list
        study["names"].append(original_name)
        del study["name"]  # delete the obsolete key

        # Return the transformed dictionary.
        return study


class Migrator_from_7_7_2_to_7_8_0(MigratorBase):
    """Migrates data from schema 7.7.2 to 7.8.0"""

    def __init__(self) -> None:
        """Invokes parent constructor and populates collection-to-transformations map."""

        super().__init__()

        # Populate the "collection-to-transformers" map for this specific migration.
        self.agenda = dict(
            study_set=[self.replace_doi_field_with_award_dois_list_field],
        )

    def replace_doi_field_with_award_dois_list_field(self, study: dict) -> dict:
        logger.info(f"Starting migration of {study['id']}")
        if "doi" in study:
            match = re.search(doi_url_pattern, study["doi"]['has_raw_value'])
            if match:
                start_index = match.end()
                as_curie = f"doi:10.{study['doi']['has_raw_value'][start_index:]}"
                study["award_dois"] = [as_curie]
            del study["doi"]
        return study


class Migrator_from_7_8_0_to_8_0_0(MigratorBase):
    """Migrates data from schema 7.8.0 to 8.0.0"""

    def __init__(self) -> None:
        """Invokes parent constructor and populates collection-to-transformations map."""

        super().__init__()

        # Populate the "collection-to-transformers" map for this specific migration.
        self.agenda = dict(
            biosample_set=[
                self.standardize_letter_casing_of_gold_biosample_identifiers],
            extraction_set=[self.rename_sample_mass_field],
            omics_processing_set=[
                self.standardize_letter_casing_of_gold_sequencing_project_identifiers],
            study_set=[self.standardize_letter_casing_of_gold_study_identifier],
        )

    def rename_sample_mass_field(self, extraction: dict) -> dict:
        logger.info(f"Starting migration of {extraction['id']}")
        if "sample_mass" in extraction:
            extraction['input_mass'] = extraction.pop('sample_mass')
        return extraction

    def standardize_letter_casing_of_gold_identifiers(self, identifiers: List[str]) -> List[str]:
        """
        Replaces the prefix `GOLD:` with `gold:` in the list of identifiers.

        Note: If the identifier contains more than one colon, everything after the second
              colon will be discarded.
        """

        standardized_identifiers = []
        for identifier in identifiers:
            logger.info(f"Original identifier: {identifier}")
            curie_parts = identifier.split(':')
            curie_prefix = curie_parts[0]  # everything before the first colon
            # everything after the first colon, assuming there are no more colons
            curie_local_id = curie_parts[1]

            # Note: `s.split(":", maxsplit=1)` could be used to divide the string at the
            #       _first_ colon only, so that `parts[1]` contains everything after that.
            #       See: https://docs.python.org/3/library/stdtypes.html#str.split

            if curie_prefix == "GOLD":
                standardized_identifiers.append(f"gold:{curie_local_id}")
            else:
                standardized_identifiers.append(identifier)

        return standardized_identifiers

    def standardize_letter_casing_of_gold_biosample_identifiers(self, biosample: dict) -> dict:
        field_name = "gold_biosample_identifiers"
        if field_name in biosample and biosample[field_name]:
            biosample[field_name] = self.standardize_letter_casing_of_gold_identifiers(
                biosample[field_name])
        else:
            biosample[field_name] = []
        return biosample

    def standardize_letter_casing_of_gold_sequencing_project_identifiers(self, omics_processing: dict) -> dict:
        field_name = "gold_sequencing_project_identifiers"
        if field_name in omics_processing and omics_processing[field_name]:
            omics_processing[field_name] = self.standardize_letter_casing_of_gold_identifiers(
                omics_processing[field_name])
        else:
            omics_processing[field_name] = []
        return omics_processing

    def standardize_letter_casing_of_gold_study_identifier(self, study: dict) -> dict:
        field_name = "gold_study_identifiers"
        if field_name in study and study[field_name]:
            study[field_name] = self.standardize_letter_casing_of_gold_identifiers(
                study[field_name])
        else:
            study[field_name] = []
        return study


class Migrator_from_8_0_to_8_1(MigratorBase):
    """previously: Migrates data from schema 8.0.0 to 8.1.0"""

    def __init__(self) -> None:
        """Invokes parent constructor and populates collection-to-transformations map."""

        super().__init__()

        # Populate the "collection-to-transformers" map for this specific migration.
        self.agenda = dict(
            study_set=[self.force_research_study_study_category],
        )

    def force_research_study_study_category(self, study: dict) -> dict:
        if 'study_category' not in study:
            logger.info(
                f"Forcing 'study_category: research_study' on {study['id']}")
            study['study_category'] = 'research_study'
        return study


class Migrator_from_8_1_to_9_0(MigratorBase):
    """previously: Migrates data from schema 8.1.0 to 9.0.0"""

    def __init__(self) -> None:
        """Invokes parent constructor and populates collection-to-transformations map."""

        super().__init__()

        # Populate the "collection-to-transformers" map for this specific migration.
        self.agenda = dict(
            study_set=[self.fix_award_dois, self.fix_pub_dois,
                       self.fix_massive, self.fix_ess_dive, self.remove_doi_slots],
        )

    def process_doi(self, study: dict, doi_list: list, doi_category: str):
        id_value = study['id']
        for doi_updates in doi_list:
            if id_value == doi_updates['id']:
                new_doi = {
                    'doi_value': doi_updates['doi'],
                    'doi_category': doi_category,
                    'doi_provider': doi_updates['doi_prov']
                }
                study.setdefault('associated_dois', []).append(new_doi)

    def fix_award_dois(self, study: dict):
        """Moves current DOIs in the award_dois slot to the new associated_dois slot.
         It also changes some DOIs that were incorrectly labeled as award to dataset DOIs.
         It also add three new award DOIs from JGI"""

        study_doi_data = load_yaml_file(
            filename='assets/misc/study_dois_changes.yaml')

        self.process_doi(
            study, study_doi_data['award_move_to_data'], 'dataset_doi')
        self.process_doi(study, study_doi_data['new_award_dois'], 'award_doi')
        self.process_doi(study, study_doi_data['award_doi_prov'], 'award_doi')

        return study

    def fix_pub_dois(self, study: dict):
        """Move publication_dois values to new associated_dois slot"""

        study.setdefault('associated_dois', []).extend(
            {'doi_value': pub_doi, 'doi_category': 'publication_doi'}
            for pub_doi in study.get('publication_dois', [])
        )

        return study

    def fix_massive(self, study: dict):
        """Change the one massive_study_identifiers value to a doi and move under new associated_dois slot"""

        mass_id = 'MASSIVE:MSV000090886'
        mass_doi = 'doi:10.25345/C58K7520G'
        study.setdefault('associated_dois', []).extend(
            {'doi_value': mass_doi, 'doi_category': 'dataset_doi', 'doi_provider': 'massive'}
            for id in study.get('massive_study_identifiers', []) if id == mass_id)

        # remove the massive_study_identifiers slot if the id matches the one to be removed in associated_dois slot
        for doi_group in study['associated_dois']:
            if doi_group['doi_value'] == mass_doi:
                study.pop('massive_study_identifiers', None)

        return study

    def fix_ess_dive(self, study: dict):
        """Move ess_dive_datasets values to associated_dois slot"""

        study.setdefault('associated_dois', []).extend(
            {'doi_value': dataset_doi, 'doi_category': 'dataset_doi',
             'doi_provider': 'ess_dive'}
            for dataset_doi in study.get('ess_dive_datasets', []))

    def remove_doi_slots(self, study: dict):
        """Remove slots that are no longer needed because their values have been moved to the associated_dois slot"""

        removal_slots = ['publication_dois', 'dataset_dois',
                         'award_dois', 'ess_dive_datasets', 'massive_study_identifiers']

        # Get dois from associated_dois
        associated_dois = [entry['doi_value']
                           for entry in study.get('associated_dois', [])]

        # Remove the old dois from the old doi slots
        for slot_name in removal_slots:
            study[slot_name] = [doi for doi in study.get(
                slot_name, []) if doi not in associated_dois]

            # Remove old doi slots if values are empty
            if not study[slot_name]:
                del study[slot_name]
            else:
                logger.error(
                    f'ERROR: Unexpected value in {slot_name} of {study["id"]} skipping slot deletion')

        # Remove associated_dois if empty (no dois were moved over and the slot is unnecessary)
        if not study['associated_dois']:
            del study['associated_dois']

        return study


@click.command()
@click_log.simple_verbosity_option(logger)
@click.option("--schema-path", default='nmdc_schema/nmdc_schema_accepting_legacy_ids.yaml', required=True, type=str,
              help="Path to the schema file to which the input YAML data file conforms")
@click.option("--input-path", default='local/mongo_as_unvalidated_nmdc_database.yaml', required=True, type=str,
              help="Path to the input YAML data file")
@click.option("--output-path", default='local/rdf_safe.yaml', required=True, type=str,
              help="Path to the output YAML data file")
@click.option("--salvage-prefix", required=True, type=str,
              help="A prefix, defined in the schema, to force for each value that the schema indicates is a CURIE but that has no prefix")
@click.option("--migrator-name", type=str, multiple=True,
              help="The name of a MigratorBase class that should be used to migrate the data")
def main(schema_path, input_path, output_path, salvage_prefix, migrator_name):
    """
    Generates a data file that conforms to a different schema version than the input data file does.

    See source code for initial and final schema versions.
    """

    migrators = []

    for current_name in migrator_name:
        try:
            migrator_class = globals()[current_name]
            migrators.append(migrator_class)
            logger.info(f"Will perform {current_name} migration")
        except KeyError:
            logger.error(
                f"ERROR: {current_name} is not a valid migrator class name")
            continue

    if len(migrators) == 0:
        logger.info("No valid migrators specified. Will perform default migration")
        exit()

    # Load the input data
    logger.info(f"Loading data from {input_path}")
    total_dict = load_yaml_file(input_path)

    # Load the schema and determine which of its slots we can migrate.
    logger.info(f"Loading schema from {schema_path}")
    view = SchemaView(schema_path)
    slots = view.all_slots()
    curie_slots = set()
    for sk, sv in slots.items():
        s_range = sv.range
        if s_range == "uriorcurie":
            slot_descendants = view.slot_descendants(sk)
            for i in slot_descendants:
                curie_slots.add(i)
        elif s_range:
            range_type_name = type(view.get_element(s_range)).class_name
            if range_type_name == "class_definition":
                if view.get_identifier_slot(s_range):
                    slot_descendants = view.slot_descendants(sk)
                    for i in slot_descendants:
                        curie_slots.add(i)

    logger.info(curie_slots)

    curie_migrator = MigratorBase()
    curie_migrator.forced_prefix = salvage_prefix

    # iterate over collections, applying CURIe fixes
    for tdk, tdv in total_dict.items():
        logger.info(f"Fixing CURIes in {tdk} if necessary")

        total_dict[tdk] = curie_migrator.fix_curies_recursively_by_key(
            tdv, set(curie_slots))

    # iterate over migrators
    for current_migrator in migrators:
        migrator = current_migrator()
        migrator.forced_prefix = salvage_prefix

        # iterate over collections, applying migration-specific transformations
        for tdk, tdv in total_dict.items():
            # If the migration specifies any transformers for this collection,
            # apply them—in order—to each document within this collection.
            transformers = migrator.get_transformers_for(collection_name=tdk)
            if len(transformers) > 0:
                logger.info(f"Starting {tdk}-specific transformations with {type(migrator)}")
                for document in tdv:
                    for transformer in transformers:
                        transformer(document)  # modifies the document in place

    # all migrations complete. save data.
    logger.info(f"Saving migrated data to {output_path}")
    with open(output_path, "w") as f:
        yaml.dump(total_dict, f)


if __name__ == "__main__":
    main()
